try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from .plugin_manager import generate_script
from .url import get_url, check_valid_url, check_valid_port


class HttpRequestTranslator(object):

    """Main Interface for the tool."""

    def __init__(self, languages=['bash'], request=None, proxy=None, search_string='', data=None):
        """Initialises all the parameters of the object.

        :param list languages: list of languages in which request's code is to be generated.
        :param str request: raw request, for which code has to be generated
        :param str proxy: custom proxy, if required in the code.
        :param str search_string: search phrase(can be regex too) to be searched in the response.
        :param str data: data string to be sent along with the header.
        """
        self.languages = languages
        self.request = request
        self.data = data
        self.proxy = proxy
        self.search_string = search_string

        # extract headers, other details(data, method, host, etc.)
        self._extract_request_details()

    def _extract_request_details(self):
        self.headers, self.details = self._parse_request()

        if self.data:
            self.details['data'] = self.data

        if self.proxy:
            # If proxy already doesn't starts with http and is like 127.0.0.1:8010
            if not self.proxy.startswith(('http', 'https')):
                proxy = get_url(self.proxy)  # Fix proxy to add appropriate scheme
            else:
                proxy = self.proxy.strip()
            try:
                self.details['proxy_host'], self.details['proxy_port'] = proxy.rsplit(":", 1)
            except ValueError:  # Missing value when splitting
                raise ValueError("Proxy provided is invalid.")
            if not check_valid_url(self.details['proxy_host']) or not check_valid_port(self.details['proxy_port']):
                raise ValueError("Proxy provided is invalid.")

    def generate_code(self):
        """Generates code for all the languages defined in the object.

        :return: A dictionary of language name and respective code.
        :rtype: dict
        """
        self._parse_request()
        all_code = {}
        for language in self.languages:
            all_code[language] = generate_script(language, self.headers, self.details, self.search_string)
        return all_code

    def _parse_request(self):
        """Parses Raw HTTP request into separate dictionaries for headers and body and other parameters.

        :param str request: Raw HTTP request.

        :raises ValueError: When request passed in malformed.

        :return: A tuple of two dictionaries where the first one is the headers and the second the details.
        :rtype: tuple
        """
        headers_lines = self.request.splitlines()
        if not headers_lines:
            raise ValueError("Request Malformed. Please Enter a Valid HTTP request.")
        new_request_method = headers_lines.pop(0)
        # Headers
        host = ''
        header_list = []
        while headers_lines:
            line = headers_lines.pop(0)
            if not line.strip('\r\n'):  # Empty line? Therefore the headers are over and the content is starting.
                break
            header_list.append(line)
            try:
                header, value = line.split(":", 1)
                # stripping one left blank space from value induced after split.
                if value.startswith(' '):
                    value = value[1:]
            except ValueError:
                raise ValueError("Headers Malformed. Please Enter a Valid HTTP request.")
            if header.lower() == "host":
                host = value.strip()  # Keep hostname for further checks
        # Data
        data = ''
        if headers_lines:
            data = ''.join(headers_lines)
        # Details
        details_dict = {}
        details_dict['data'] = data
        details_dict['method'] = new_request_method.split(' ', 1)[0].strip()
        details_dict['Host'] = host
        # Not using whatever stored in parsed_request for the reason to keep the request as original as possible
        try:  # try to split the path from request if one is passed.
            proto_ver = new_request_method.split(' ', 2)[2].split('/', 1)
            details_dict['protocol'] = proto_ver[0].strip()
            details_dict['version'] = proto_ver[1].strip()
            details_dict['path'] = new_request_method.split(' ', 2)[1].strip()
        except IndexError:
            details_dict['path'] = ""
            try:
                proto_ver = new_request_method.split(' ', 2)[1].split('/', 1)
            except IndexError:  # Failed to get protocol and version.
                raise ValueError("Request Malformed. Please Enter a Valid HTTP request.")
            details_dict['protocol'] = proto_ver[0].strip()
            details_dict['version'] = proto_ver[1].strip()
        # Parse the GET Path to update it to only contain the relative path and not whole url
        # scheme://netloc/path;parameters?query#fragment
        # Eg: Path=https://google.com/robots.txt to /robots.txt
        scheme, netloc, path, params, query, frag = urlparse(details_dict['path'])
        if params:
            path = path + ";" + params
        if query:
            path = path + "?" + query
        if frag:
            path = path + "#" + frag
        details_dict['path'] = path
        # If scheme is specified in GET Path and Header 'Host' Field doesn't already starts with it
        if scheme and not host.startswith(scheme):
            details_dict['pre_scheme'] = scheme + "://"  # Store the scheme defined in GET path for later checks
        else:
            details_dict['pre_scheme'] = ''

        return header_list, details_dict




try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from .util import re_ipv4_address, re_ipv6_address, re_domain


def check_valid_url(url):
    """Verify that a URL (containing the protocol) is valid.

    :param str url: URL containing the protocol to validate.

    :return: ``True`` if `url` is valid, ``False`` otherwise.
    :rtype: bool
    """
    _, netloc, _, _, _, _ = urlparse(url)
    if not netloc:  # No protocol specified and urlparse assumed it was a relative path.
        return False
    if not netloc.startswith('['):  # URL without IPv6 e.g. [::1]
        if ':' in netloc:
            netloc, port = netloc.rsplit(':', 1)
            if port and not check_valid_port(port):
                return False
        if re_ipv4_address.match(netloc):
            return True
        if re_domain.match(netloc):
            return True
    else:
        if not netloc.endswith(']'):  # Is there a port specified? e.g. [::1]:443
            netloc, port = netloc.rsplit(':', 1)
            if port and not check_valid_port(port):
                return False
        if re_ipv6_address.match(netloc[1:-1]):
            return True
    return False


def check_valid_port(port):
    """Verify that a port is valid.

    :param str port: port to validate.

    :return: ``True`` if `port` is valid, ``False`` otherwise.
    :rtype: bool
    """
    try:
        port = int(port)
    except ValueError:
        return False
    if port < 0 or port > 65535:
        return False
    return True


def get_url(host, pre_protocol=None):
    """Generate URL based on the host.

    :param str host: Host from which to generate the URL (e.g. google.com:443).
    :param str pre_protocol: Protocol passed in the GET Path. Example request:
        GET https://127.0.0.1/robots.txt HTTP/1.1
        Host: 127.0.0.1:323
        Then, pre_protocol will be `https`.
        Defaults to `None`.

    :return: URL with domain and protocol (e.g. https://google.com).
    :rtype: str
    """
    port_protocol = {'443': 'https', '22': 'ssh', '21': 'ftp', '20': 'ftp', '113': 'irc', '80': 'http'}
    protocol = ''
    url = host.strip()
    if not url.startswith('['):  # A port is specified in the domain and without IPV6
        if ':' in url:
            _, port = url.rsplit(':', 1)
            if port in port_protocol:  # Do we know the protocol?
                protocol = port_protocol[port] + '://'
    else:
        if not url.endswith(']'):   # IPV6 url with Port, [::1]:443
            _, port = url.rsplit(':', 1)
            if port in port_protocol:  # Do we know the protocol?
                protocol = port_protocol[port] + '://'
    # If GET path already specifies a protocol, give preference to that
    protocol = pre_protocol or protocol or 'http://'  # Default protocol set to http
    return protocol + url



from __future__ import print_function

import sys
try:
    input = raw_input  # Python 2.x
except NameError:
    pass  # Python 3.x
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from .plugin_manager import generate_script
from .url import get_url, check_valid_url


def process_arguments(args):
    """Process the arguments provided to the translator and places values in separate headers and details dictionaries.

    :param class `argparse.Namespace`: `argparse` class object containing arguments passed to the translator.

    :raises ValueError: When proxy provided is invalid.

    :return: A dictionaries of arguments passed.
    :rtype: `dict`
    """
    argdict = vars(args)
    try:
        script_list = argdict['language'][0].split(',')
    except TypeError:
        script_list = []  # If --language option is not used.
    if args.interactive:
        try:
            take_headers(script_list)
        except KeyboardInterrupt:
            print("\nThanks for using the interactive mode! Exiting!")
            sys.exit(0)
    else:
        if args.request:
            headers, details = parse_raw_request(args.request)
        elif args.file:
            fp = open(args.file)
            raw_request = ""
            for line in fp.readlines():
                if line != '':
                    raw_request += line
            raw_request = raw_request.rstrip('\r\n')  # Remove any linefeeds if present in the file provided
            headers, details = parse_raw_request(raw_request)
        else:
            print("Input a raw HTTP Request and try again.\nElse try using the interactive option")
            sys.exit(-1)
        if args.data:
            details['data'] = args.data
        if args.proxy:
            # If proxy already doesn't starts with http and is like 127.0.0.1:8010
            if not args.proxy.startswith(('http', 'https')):
                proxy = get_url(args.proxy)  # Fix proxy to add appropriate scheme
            else:
                proxy = args.proxy.strip()
            if not check_valid_url(proxy):
                raise ValueError("Proxy provided is invalid.")
            try:
                details['proxy_host'], details['proxy_port'] = proxy.rsplit(":", 1)
            except IndexError:
                raise ValueError("Proxy provided is invalid.")
        if not details['data'] and details['method'].strip().upper() == "POST":
            print("Hi there. Send some data to POST, use --data for sending data.")
            sys.exit(-1)
        arg_option = None
        if args.search_string:
            arg_option = args.search_string
        elif args.search_regex:
            arg_option = args.search_regex
        if len(script_list) == 0:
            generated_code = generate_script('bash', headers, details, arg_option)
            print(generated_code)
        else:
            for script in script_list:
                generated_code = generate_script(script, headers, details, arg_option)
                print(generated_code)

    return argdict


def take_headers(script_list):
    """Takes Request headers while interactive mode is active

    :param list script_list: List of names of the languages for which script is to be generated.

    :return: Raw Request Headers passed in the interactive session.
    :rtype: `str`
    """
    headers = []
    print("Enter request headers (Ctrl+D to finish/Ctrl+C to quit).")
    while True:
        try:
            uentered = input(">>> ")
            headers.append(uentered + "\n")
        except EOFError:
            take_body(headers, script_list)
    return headers


def take_body(headers, script_list):
    """Takes Request body while interactive mode is active.
    Also prints the generated code for languages passed in `script_list` on console.

    :param str headers: Raw request headers feeded in the interactive session.
    :param list script_list: List of names of the languages for which script is to be generated.

    :return: Body for the request passed in interactive session.
    :rtype: `str`
    """
    body = []
    print("\nEnter request body/parameters (Ctrl+D to finish/Ctrl+C to quit).")
    while True:
        try:
            uentered = input(">>> ")
            body.append(uentered + "\n")
        except EOFError:
            try:
                headers, details = parse_raw_request("".join(headers))
                details['data'] = "".join(body).strip()
                if not script_list:
                    print(generate_script('bash', headers, details))
                else:
                    for script in script_list:
                        print(generate_script(script, headers, details))
            except ValueError:
                print("Please Enter a Vaild Request!")
            take_headers(script_list)
    return body


def parse_raw_request(request):
    """Parses Raw HTTP request into separate dictionaries for headers and body and other parameters.

    :param str request: Raw HTTP request.

    :raises ValueError: When request passed in malformed.

    :return: A tuple of two dictionaries where the first one is the headers and the second the details.
    :rtype: tuple
    """
    headers_lines = request.splitlines()
    if not headers_lines:
        raise ValueError("Request Malformed. Please Enter a Valid HTTP request.")
    new_request_method = headers_lines.pop(0)
    # Headers
    header_list = []
    while headers_lines:
        line = headers_lines.pop(0)
        if not line.strip('\r\n'):  # Empty line? Therefore the headers are over and the content is starting.
            break
        header_list.append(line)
        try:
            header, value = line.split(":", 1)
        except IndexError:
            raise ValueError("Headers Malformed. Please Enter a Valid HTTP request.")
        if header.lower() == "host":
            host = value.strip()  # Keep hostname for further checks
    # Data
    data = ''
    if headers_lines:
        data = ''.join(headers_lines)
    # Details
    details_dict = {}
    details_dict['data'] = data
    details_dict['method'] = new_request_method.split(' ', 1)[0].strip()
    details_dict['Host'] = host
    # Not using whatever stored in parsed_request for the reason to keep the request as original as possible
    try:  # try to split the path from request if one is passed.
        proto_ver = new_request_method.split(' ', 2)[2].split('/', 1)
        details_dict['protocol'] = proto_ver[0].strip()
        details_dict['version'] = proto_ver[1].strip()
        details_dict['path'] = new_request_method.split(' ', 2)[1].strip()
    except IndexError:
        details_dict['path'] = ""
        try:
            proto_ver = new_request_method.split(' ', 2)[1].split('/', 1)
        except IndexError:  # Failed to get protocol and version.
            raise ValueError("Request Malformed. Please Enter a Valid HTTP request.")
        details_dict['protocol'] = proto_ver[0].strip()
        details_dict['version'] = proto_ver[1].strip()
    # Parse the GET Path to update it to only contain the relative path and not whole url
    # scheme://netloc/path;parameters?query#fragment
    # Eg: Path=https://google.com/robots.txt to /robots.txt
    scheme, netloc, path, params, query, frag = urlparse(details_dict['path'])
    if params:
        path = path + ";" + params
    if query:
        path = path + "?" + query
    if frag:
        path = path + "#" + frag
    details_dict['path'] = path
    # If scheme is specified in GET Path and Header 'Host' Field doesn't already starts with it
    if scheme and not host.startswith(scheme):
        details_dict['pre_scheme'] = scheme + "://"  # Store the scheme defined in GET path for later checks
    else:
        details_dict['pre_scheme'] = ''
    return (header_list, details_dict)



from __future__ import print_function

from .base import AbstractScript
from .script import BashScript, PHPScript, PythonScript, RubyScript


def get_script_class(script_name):
    """Returns the class of the script.

    :param str script_name: language name of the script class which we want to import

    :raises ValueError: When the script is not supported.

    :return: The class of the `script_name` script.
    :rtype: :class:`AbstractScript`
    """
    script_name = script_name.strip().lower()

    for script_class in AbstractScript.__subclasses__():
        if script_name == script_class.__language__:
            return script_class
    raise ValueError("The {} language is not supported.".format(script_name))


def generate_script(script, headers, details, search_string=None):
    """Returns the script code for the HTTP request passed in script language

    :param str script: Name of the language for which script is to be generated
    :param dict headers: Headers information
    :param dict details: Details information
    :param str search_string: string to be searched for in the response for given request

    :return: A combined string of generated code
    :rtype: `str`
    """
    class_script = get_script_class(script.strip().lower())
    return class_script(headers=headers, details=details, search=search_string).generate_script()



"""

:synopsis: Define the specialize script classes that will generate the script code.

"""

from .base import AbstractScript


class BashScript(AbstractScript):

    """Extended `AbstractScript` class for Bash script code generation.
    Fills code variables for the request from `bash_template`.
    Overrides `_generate_request` method to generate bash specific code.
    """

    __language__ = 'bash'

    def _generate_request(self):
        code = self.code_nosearch.format(
            method=self.details.get('method', ''),
            url=self.url,
            headers=self._generate_headers())
        return code


class PHPScript(AbstractScript):

    """Extended `AbstractScript` class for PHP script code generation.
    Fills code variables for the request from `php_template`.
    Overrides `_generate_begin` method to generate php specific code.
    """

    __language__ = 'php'

    def _generate_begin(self):
        return self.code_begin.format(url=self.url) + self._generate_headers()


class PythonScript(AbstractScript):

    """Extended `AbstractScript` class for Python script code generation.
    Fills code variables for the request from `python_template`.
    Overrides `_generate_begin` method to generate python specific code.
    """

    __language__ = 'python'

    def _generate_begin(self):
        return self.code_begin.format(url=self.url, headers=str(self.headers))


class RubyScript(AbstractScript):

    """Extended `AbstractScript` class for Ruby script code generation.
    Fills code variables for the request from `ruby_template`.
    Overrides `_generate_begin` method to generate Ruby specific code.
    """

    __language__ = 'ruby'

    def _generate_begin(self):
        code = self.code_begin.format(url=self.url, method=self.details.get('method', '').strip().lower())
        code += self.code_headers.format(headers=self._generate_headers())
        return code
